import 'dart:convert';
import 'dart:io';
import 'dart:math';

import 'package:flutter/foundation.dart';
import 'package:flutterx_http/flutterx_http.dart';

@immutable
class MultipartRequest extends IRequest {
  static final RegExp _asciiOnly = RegExp(r'^[\x00-\x7F]+$');
  final String method;
  final Uri url;
  final QueryParameters params;
  final HeadersMap headers;
  final bool computeContentLength;
  final Map<String, String> fields;
  final List<MultipartFile> files;

  const MultipartRequest({
    this.method = 'POST',
    required this.url,
    this.params = const {},
    this.headers = const {},
    this.computeContentLength = false,
    this.fields = const {},
    this.files = const [],
  });

  @override
  XHttpRequest finalize() {
    final boundary = 'dart-http-boundary-${Random().nextInt(4294967296).toString().padLeft(10, '0')}';
    return XHttpRequest(
        method: method,
        url: url.appendQueryParameters(params),
        headers: {...headers, HttpHeaders.contentTypeHeader: 'multipart/form-data; boundary=$boundary'},
        body: XHttpContent.fromStream(_finalize(boundary),
            contentLength: computeContentLength ? _computeLength(boundary) : null));
  }

  Stream<List<int>> _finalize(String boundary) async* {
    const line = [13, 10]; // \r\n
    final separator = utf8.encode('--$boundary\r\n');
    final close = utf8.encode('--$boundary--\r\n');
    for (final field in fields.entries) {
      yield separator;
      yield utf8.encode(_headerForField(field.key, field.value));
      yield utf8.encode(field.value);
      yield line;
    }
    for (final file in files) {
      yield separator;
      yield utf8.encode(_headerForFile(file));
      yield* file.content.asStream();
      yield line;
    }
    yield close;
  }

  int _computeLength(String boundary) {
    var length = 0;
    final boundaryLength = boundary.length;
    fields.forEach((name, value) {
      length += // '--'.length +
          boundaryLength +
              // '\r\n'.length +
              utf8.encode(_headerForField(name, value)).length +
              utf8.encode(value).length +
              // '\r\n'.length
              6;
    });
    for (final file in files)
      length += // '--'.length +
          boundaryLength +
              // '\r\n'.length +
              utf8.encode(_headerForFile(file)).length +
              file.content.contentLength! +
              // '\r\n'.length
              6;
    return length +
        // '--'.length +
        boundaryLength +
        // '--\r\n'.length
        4;
  }

  static String _headerForField(String name, String value) {
    var header = 'content-disposition: form-data; name="${MultipartFile.encodeName(name)}"';
    if (!_asciiOnly.hasMatch(value))
      header = '$header\r\n'
          'content-type: text/plain; charset=utf-8\r\n'
          'content-transfer-encoding: binary';
    return '$header\r\n\r\n';
  }

  static String _headerForFile(MultipartFile file) {
    var header = 'content-type: ${file.contentType}\r\n'
        'content-disposition: form-data; name="${MultipartFile.encodeName(file.field)}"';
    final filename = file.filename;
    if (filename != null) header = '$header; filename="${MultipartFile.encodeName(filename)}"';
    return '$header\r\n\r\n';
  }
}
